# Copyright 2020, castLabs GmbH

from os import environ

################################################################################

class _ForwardError_(Exception):
  def __init__(self, cls, msg):
    self._forward_class_name_ = cls
    super().__init__(msg)

class _ForwardArgs_:
  def __init__(self, fn, *args, **kwargs):
    self._fn_ = fn
    self._args_ = args
    self._kwargs_ = kwargs
  def __call__(self, *args, **kwargs):
    return self._fn_(*self._args_, *args, **self._kwargs_, **kwargs)

def conclude(e):
  j = {
    'errorType': getattr(e, '_forward_class_name_', e.__class__.__name__),
    'errorMessage': str(e),
  }
  if e.__cause__:
    j['errorCause'] = conclude(e.__cause__)
  return j

def gather(j):
  from . import error
  import builtins
  t = j.get('errorType', 'UnknownError')
  c = getattr(error, t, getattr(builtins, t, _ForwardArgs_(_ForwardError_, t)))
  e = c(j.get('errorMessage', j.get('message', 'No message')))
  if 'errorCause' in j:
    e.__cause__ = gather(j['errorCause'])
  return e

################################################################################

def die(e, code=1):
  if 'EVS_TRACE' in environ:
    raise e
  print(f'{getattr(e, "_forward_class_name_", e.__class__.__name__)}: {e}')
  e = e.__cause__
  while e:
    n, m = getattr(e, '_forward_class_name_', e.__class__.__name__), str(e)
    if m:
      print(f' <- {n}: {m}')
    else:
      print(f' <- {n}')
    e = e.__cause__
  raise SystemExit(code)

################################################################################



# Copyright 2020, castLabs GmbH

from . import __version__
from .core.error import NonInteractiveError

from sys import stdin
from os import environ, path, makedirs
import errno
from getpass import getpass
import json

import boto3
from botocore.config import Config

################################################################################

DEFAULT_CONFIG_FILE = '~/.config/evs/config.json'
DEFAULT_CLIENT_ID = '4mdbpdvnjc0tq0lvi96aqvictb'
DEFAULT_CLIENT_REGION = 'us-east-1'

################################################################################

class Auth:
  DEFAULT_CONNECT_TIMEOUT = 60
  DEFAULT_AUTH_TIMEOUT = 60

  def __init__(self, no_refresh=False, no_ask=True, config_file=None, client_id=None, region=None,
               connect_timeout=None, auth_timeout=None, **_):
    self._version = __version__
    self._config_file = path.expanduser(config_file or environ.get('EVS_CONFIG_FILE', DEFAULT_CONFIG_FILE))
    self._config_dir = path.dirname(self._config_file)
    self._client_id = client_id or environ.get('EVS_CLIENT_ID', DEFAULT_CLIENT_ID)
    self._client_region = region or environ.get('EVS_CLIENT_REGION', DEFAULT_CLIENT_REGION)
    self._connect_timeout = float(connect_timeout or environ.get('EVS_CONNECT_TIMEOUT', self.DEFAULT_CONNECT_TIMEOUT))
    self._auth_timeout = float(auth_timeout or environ.get('EVS_AUTH_TIMEOUT', self.DEFAULT_AUTH_TIMEOUT))
    self._auth_account_name = None
    self._auth_passwd = None
    self._auth_access_token = None
    self._auth_refresh_token = None
    self._auth_id_token = None
    self.__refreshed_auth = False
    self.__no_ask = no_ask or 'EVS_NO_ASK' in environ or not stdin.isatty()
    self._load_config()

  def _log(self, msg, end='\n'):
    print(msg, end=end, flush=True)

  def _on_load_config(self, j):
    a = j.get('Auth', {})
    account_name = a.get('AccountName')
    if self._auth_account_name and account_name != self._auth_account_name:
      return
    self._auth_account_name = account_name
    self._auth_access_token = self._auth_access_token or a.get('AccessToken')
    self._auth_refresh_token = self._auth_refresh_token or a.get('RefreshToken')
    self._auth_id_token = self._auth_id_token or a.get('IdToken')

  def _load_config(self):
    try:
      with open(self._config_file) as fh:
        self.__config = json.load(fh)
        self._on_load_config(self.__config)
    except:
      self.__config = {}

  def _on_store_config(self, j={}):
    j['Auth'] = {
      'AccountName': self._auth_account_name,
      'AccessToken': self._auth_access_token,
      'RefreshToken': self._auth_refresh_token,
      'IdToken': self._auth_id_token,
    }
    return j

  def _store_config(self):
    if not self._config_file:
      return False
    try:
      try:
        makedirs(self._config_dir)
      except OSError as e:
        if errno.EEXIST != e.errno or not path.isdir(self._config_dir):
          return False
      with open(self._config_file, 'w') as fh:
        json.dump(self._on_store_config(self.__config), fh, indent=2)
      return True
    except Exception as e:
      return False

  def _boto_session(self):
    if not hasattr(self, '_Auth__boto_session'):
      for k in environ.keys():
        if k.startswith('AWS_'):
          del environ[k]
      self.__boto_session = boto3.Session(region_name=self._client_region)
    return self.__boto_session

  def _auth_client(self):
    if not hasattr(self, '_Auth__auth_client'):
      self.__auth_client = self._boto_session().client(
        'cognito-idp',
        config=Config(connect_timeout=self._connect_timeout, read_timeout=self._auth_timeout),
      )
    return self.__auth_client

  def _ask(self, title, default=None, min=None, max=None, secret=False, confirm=False):
    if self.__no_ask:
      if confirm:
        raise NonInteractiveError(f'\'{title}\' attribute requires user confirmation')
      if default is None:
        raise NonInteractiveError(f'\'{title}\' attribute has no default value')
      return default
    default = default or ''
    prompt = f'>> {title} [{"*" * len(default) if secret else default}]: '
    fn = getpass if secret else input
    while True:
      v = fn(prompt) or default
      if min is not None and len(v) < min:
        self._log(f' - \'{title}\' needs to be at least {min} characters long')
        continue
      if max is not None and len(v) > max:
        self._log(f' - \'{title}\' needs to be at most {max} characters long')
        continue
      return v

  def _refresh_auth(self, account_name=None, passwd=None):
    if account_name and account_name != self._auth_account_name:
      self._clear_auth()
    if self.__refreshed_auth:
      return False
    self.__refreshed_auth = True
    self._auth_access_token = None
    try:
      client = self._auth_client()
      if self._auth_refresh_token:
        try:
          r = client.initiate_auth(
            ClientId=self._client_id,
            AuthFlow='REFRESH_TOKEN_AUTH',
            AuthParameters={'REFRESH_TOKEN': self._auth_refresh_token}
          )
          j = r['AuthenticationResult']
          self._auth_access_token = j['AccessToken']
          self._auth_id_token = j['IdToken']
        except Exception as e:
          self._auth_refresh_token = self._auth_id_token = None
      if not self._auth_access_token:
        try:
          self._auth_account_name = account_name or self._ask('Account Name', default=self._auth_account_name, min=1)
          self._auth_passwd = passwd or self._ask('Password', default=self._auth_passwd, min=1, secret=True)
          r = client.initiate_auth(
            ClientId=self._client_id,
            AuthFlow='USER_PASSWORD_AUTH',
            AuthParameters={'USERNAME': self._auth_account_name, 'PASSWORD': self._auth_passwd}
          )
          j = r['AuthenticationResult']
          self._auth_access_token = j['AccessToken']
          self._auth_refresh_token = j['RefreshToken']
          self._auth_id_token = j['IdToken']
        except Exception as e:
          self._auth_access_token = self._auth_refresh_token = self._auth_id_token = None
          raise
      return True
    finally:
      self._store_config()

  def _clear_auth(self):
    self._auth_access_token = None
    self._auth_refresh_token = None
    self._auth_id_token = None
    self.__refreshed_auth = False

  def _check_auth(self, account_name=None, passwd=None):
    account_name = account_name or self._auth_account_name
    if account_name != self._auth_account_name or not self._auth_access_token:
      self._clear_auth()
      return self._refresh_auth(account_name, passwd)
    return False

################################################################################



# Copyright 2020, castLabs GmbH

from .. import __version__
from ..core.failure import die
from ..account import Account
from argparse import ArgumentParser
from sys import argv
from os import environ, path

################################################################################

def main():
  try:
    bp = ArgumentParser(path.basename(argv[0]))
    bp.add_argument('-v', '--version', action='version', version=f'%(prog)s {__version__}')
    bp.add_argument('-n', '--no-ask', help='enable non-interactive mode',
      action='store_true', default='EVS_NO_ASK' in environ,
    )

    _timeout_grp = bp.add_argument_group('timeout options')
    _timeout_grp.add_argument('--connect-timeout',
      help=f'connection timeout in seconds (default: {Account.DEFAULT_CONNECT_TIMEOUT})',
      type=float, default=environ.get('EVS_CONNECT_TIMEOUT', Account.DEFAULT_CONNECT_TIMEOUT)
    )
    _timeout_grp.add_argument('--auth-timeout',
      help=f'auth request timeout in seconds (default: {Account.DEFAULT_AUTH_TIMEOUT})',
      type=float, default=environ.get('EVS_AUTH_TIMEOUT', Account.DEFAULT_AUTH_TIMEOUT)
    )

    _cred = ArgumentParser(add_help=False)
    _cred.add_argument('-A', '--account-name', help='provide account name [CACHED]',
      default=environ.get('EVS_ACCOUNT_NAME')
    )
    _cred.add_argument('-P', '--passwd', help='provide account password',
      default=environ.get('EVS_PASSWD')
    )
    _email = ArgumentParser(add_help=False)
    _email.add_argument('-E', '--email', help='provide email address [CACHED]',
      default=environ.get('EVS_EMAIL')
    )
    _details = ArgumentParser(add_help=False)
    _details.add_argument('-F', '--first-name', help='provide first name [CACHED]',
      default=environ.get('EVS_FIRST_NAME')
    )
    _details.add_argument('-L', '--last-name', help='provide last name [CACHED]',
      default=environ.get('EVS_LAST_NAME')
    )
    _details.add_argument('-O', '--organization', help='provide organization [CACHED]',
      default=environ.get('EVS_ORGANIZATION')
    )
    _no_confirm = ArgumentParser(add_help=False)
    _no_confirm.add_argument('-N', '--no-confirm', help='don\'t automatically ask for confirmation',
      default=environ.get('EVS_NO_CONFIRM')
    )
    _code = ArgumentParser(add_help=False)
    _code.add_argument('-C', '--code', help='provide confirmation code',
      default=environ.get('EVS_CODE')
    )

    sp = bp.add_subparsers(title='commands', dest='command', required=True)
    cp = sp.add_parser('signup', help='sign up for EVS account',
      aliases=['sup'],
      parents=[_cred, _email, _details, _no_confirm]
    )
    cp.set_defaults(command=Account.signup)
    cp = sp.add_parser('resend', help='resend EVS account confirmation code',
      aliases=['rs'],
      parents=[_cred, _no_confirm]
    )
    cp.set_defaults(command=Account.resend)
    cp = sp.add_parser('confirm-signup', help='confirm account signup',
      aliases=['csup'],
      parents=[_cred, _code]
    )
    cp.set_defaults(command=Account.confirm_signup)
    cp = sp.add_parser('update', help='update account details',
      aliases=['up'],
      parents=[_cred, _details]
    )
    cp.set_defaults(command=Account.update)
    cp = sp.add_parser('reset', help='reset account password',
      aliases=['res'],
      parents=[_cred, _no_confirm]
    )
    cp.set_defaults(command=Account.reset)
    cp = sp.add_parser('confirm-reset', help='confirm account password reset',
      aliases=['cres'],
      parents=[_cred, _code]
    )
    cp.set_defaults(command=Account.confirm_reset)
    cp = sp.add_parser('refresh', help='refresh authorization tokens',
      aliases=['r'],
      parents=[_cred]
    )
    cp.set_defaults(command=Account.refresh)
    cp = sp.add_parser('deauth', help='discard any authorization tokens',
      aliases=['da'],
      parents=[]
    )
    cp.set_defaults(command=Account.deauth)
    cp = sp.add_parser('reauth', help='discard any authorization tokens & refresh',
      aliases=['ra'],
      parents=[_cred]
    )
    cp.set_defaults(command=Account.reauth)
    cp = sp.add_parser('delete', help='delete EVS account',
      aliases=['del'],
      parents=[_cred]
    )
    cp.set_defaults(command=Account.delete)

    args = vars(bp.parse_args())
    account = Account(no_ask=args.pop('no_ask'),
                      connect_timeout=args.pop('connect_timeout', None),
                      auth_timeout=args.pop('auth_timeout', None))
    run = args.pop('command')
    run(account, **args)
  except (Exception, KeyboardInterrupt) as e:
    die(e)

################################################################################



# Copyright 2020, castLabs GmbH

from .auth import Auth

################################################################################

class Account(Auth):
  def __init__(self, no_ask=True, **_):
    self._account_name = None
    self._first_name = None
    self._last_name = None
    self._email = None
    self._organization = None
    super().__init__(no_ask=no_ask)

  def _on_load_config(self, j):
    super()._on_load_config(j)
    i = j.get('Account', {})
    account_name = i.get('AccountName')
    if self._account_name and account_name != self._account_name:
      return
    self._account_name = account_name
    self._first_name = self._first_name or i.get('FirstName')
    self._last_name = self._last_name or i.get('LastName')
    self._email = self._email or i.get('Email')
    self._organization = self._organization or i.get('Organization')

  def _on_store_config(self, j={}):
    j['Account'] = {
      'FirstName': self._first_name,
      'LastName': self._last_name,
      'Email': self._email,
      'Organization': self._organization,
      'AccountName': self._account_name,
    }
    return super()._on_store_config(j)

  def _refresh_account(self, account_name=None, passwd=None):
    client = self._auth_client()
    while True:
      try:
        r = client.get_user(
          AccessToken=self._auth_access_token,
        )
        r = { i['Name']: i['Value'] for i in r['UserAttributes'] }
        self._email = r.get('email')
        self._first_name = r.get('given_name', self._first_name)
        self._last_name = r.get('family_name', self._first_name)
        self._organization = r.get('custom:organization', self._organization)
        self._account_name = self._auth_account_name
        self._store_config()
        break
      except client.exceptions.NotAuthorizedException as e:
        if not self._refresh_auth():
          raise

  def signup(self, account_name=None, passwd=None, email=None, first_name=None, last_name=None, organization=None, no_confirm=False, **_):
    self._log('Signing up for castLabs EVS')
    self._log(' - A valid e-mail address is required for account verification')
    self._email = email or self._ask('E-mail Address', default=self._email, min=5, confirm=True)
    self._first_name = first_name or self._ask('First Name', default=self._first_name, min=1, confirm=True)
    self._last_name = last_name or self._ask('Last Name', default=self._last_name, min=1, confirm=True)
    self._organization = organization or self._ask('Organization', default=self._organization, confirm=True)
    self._account_name = account_name or self._ask('Account Name', default=self._account_name or self._auth_account_name, min=1, confirm=True)
    if not passwd:
      while True:
        passwd = self._ask('Password', min=8, secret=True)
        if passwd == self._ask('Verify Password', min=8, secret=True):
          break
        self._log(' - Passwords do not match')
    self._store_config()
    r = self._auth_client().sign_up(
      ClientId=self._client_id,
      Username=self._account_name,
      Password=passwd,
      UserAttributes=[
        {'Name': 'email', 'Value': self._email},
        {'Name': 'given_name', 'Value': self._first_name},
        {'Name': 'family_name', 'Value': self._last_name},
        {'Name': 'custom:organization', 'Value': self._organization},
      ]
    )
    if not r['UserConfirmed'] and not no_confirm:
      self.confirm_signup(self._account_name, passwd)

  def resend(self, account_name=None, passwd=None, no_confirm=False, **_):
    self._log('Resending EVS account confirmation code')
    account_name = account_name or self._ask('Account Name', default=self._account_name or self._auth_account_name, min=1, confirm=True)
    r = self._auth_client().resend_confirmation_code(
      ClientId=self._client_id,
      Username=self._account_name
    )
    if not no_confirm:
      self.confirm_signup(self._account_name, passwd)

  def confirm_signup(self, account_name=None, passwd=None, code=None, **_):
    self._log('Confirming EVS account')
    self._log(' - A confirmation code has been sent to your e-mail address')
    account_name = account_name or self._ask('Account Name', default=self._account_name or self._auth_account_name, min=1, confirm=True)
    code = code or self._ask('Confirmation Code', min=6)
    r = self._auth_client().confirm_sign_up(
      ClientId=self._client_id,
      Username=account_name,
      ConfirmationCode=code,
    )
    self.reauth(account_name, passwd)

  def update(self, account_name=None, passwd=None, first_name=None, last_name=None, organization=None):
    self._log('Updating EVS account details')
    account_name = account_name or self._ask('Account Name', default=self._auth_account_name or self._account_name, min=1, confirm=True)
    if self._check_auth(account_name, passwd):
      self._refresh_account(account_name, passwd)
    self._first_name = first_name or self._ask('First Name', default=self._first_name, min=1, confirm=True)
    self._last_name = last_name or self._ask('Last Name', default=self._last_name, min=1, confirm=True)
    self._organization = organization or self._ask('Organization', default=self._organization, confirm=True)
    self._store_config()
    client = self._auth_client()
    while True:
      try:
        r = client.update_user_attributes(
          AccessToken=self._auth_access_token,
          UserAttributes=[
            {'Name': 'given_name', 'Value': self._first_name},
            {'Name': 'family_name', 'Value': self._last_name},
            {'Name': 'custom:organization', 'Value': self._organization},
          ]
        )
        break
      except client.exceptions.NotAuthorizedException as e:
        if not self._refresh_auth():
          raise

  def reset(self, account_name=None, passwd=None, no_confirm=False, **_):
    self._log('Requesting EVS account password reset')
    self._log(' - A confirmation code will be sent to your e-mail address')
    account_name = account_name or self._ask('Account Name', default=self._auth_account_name or self._account_name, min=1, confirm=True)
    r = self._auth_client().forgot_password(
      ClientId=self._client_id,
      Username=account_name,
    )
    if not no_confirm:
      self.confirm_reset(account_name, passwd)

  def confirm_reset(self, account_name=None, passwd=None, code=None, **_):
    self._log('Confirming EVS account password reset')
    self._log(' - A confirmation code has been sent to your e-mail address')
    account_name = account_name or self._ask('Account Name', default=self._auth_account_name or self._account_name, min=1, confirm=True)
    if not passwd:
      while True:
        passwd = self._ask('New Password', min=8, secret=True)
        if passwd == self._ask('Verify Password', min=8, secret=True):
          break
        self._log(' - Passwords do not match')
    code = code or self._ask('Confirmation Code', min=6)
    r = self._auth_client().confirm_forgot_password(
      ClientId=self._client_id,
      Username=account_name,
      Password=passwd,
      ConfirmationCode=code,
    )
    self.reauth(account_name, passwd)

  def refresh(self, account_name=None, passwd=None, **_):
    self._log("Refreshing authorization token(s)")
    if self._refresh_auth(account_name, passwd):
      self._refresh_account(account_name, passwd)

  def deauth(self, **_):
    self._log("Discarding authorization token(s)")
    self._clear_auth()
    self._store_config()

  def reauth(self, account_name=None, passwd=None, **_):
    self.deauth()
    self.refresh(account_name, passwd)

  def delete(self, account_name=None, passwd=None, **_):
    self._log('Deleting EVS account')
    self._log(' - Account deletion requires user confirmation')
    account_name = self._ask('Confirm Account Name', default=account_name or self._auth_account_name or self._account_name, min=1, confirm=True)
    if self._check_auth(account_name, passwd):
      self._refresh_account(account_name, passwd)
    client = self._auth_client()
    while True:
      try:
        r = client.delete_user(
          AccessToken=self._auth_access_token
        )
        break
      except client.exceptions.NotAuthorizedException as e:
        if not self._refresh_auth():
          raise

################################################################################

if __name__ == '__main__':
  from .cli.account import main
  main()

################################################################################



# Copyright 2020, castLabs GmbH

from .core.error import *
from .core.util import detect_bin, DEFAULT_HASH_BLOCK, hash_macho_wv0, hash_pe_wv0
from .core.failure import gather
from .auth import Auth

from sys import stdout
from io import BytesIO
from os import environ, path, rename, remove, makedirs
from time import monotonic
from math import ceil
import errno
from datetime import datetime, timedelta, timezone
from binascii import a2b_base64, b2a_base64
import json
from glob import glob
from zlib import compressobj, MAX_WBITS

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.hashes import SHA1
from cryptography.x509 import load_der_x509_certificate, ObjectIdentifier
from cryptography.x509.oid import ExtensionOID, ExtendedKeyUsageOID
from cryptography.hazmat.primitives.asymmetric.padding import PSS, MGF1

import requests
import boto3
from boto3.s3.transfer import TransferConfig, create_transfer_manager
from botocore.config import Config

################################################################################

DEFAULT_API_URL = 'https://evs-api.castlabs.com'

################################################################################

ExtensionOID.VMP_DEVELOPMENT = ObjectIdentifier('1.3.6.1.4.1.11129.4.1.2')
ExtensionOID.VMP_PERSISTENT = ObjectIdentifier('1.3.6.1.4.1.11129.4.1.3')

STREAMING_SIGNATURE='streaming'
PERSISTENT_SIGNATURE='persistent'

SUBJECT_KEY_IDENTIFIERS = {
  STREAMING_SIGNATURE: [
    b'\x0f\x02\xae\xdf\xf7\xe5\xca\x19\nK\xc4$\xdb\xfat\xb5i\x97\x18\x8c',
    b'\xf8\xa3\xcbQZ\xe1\xc4\xe45\x07"n\x9eS\x13U\xe6\x85Y*'
  ],
  PERSISTENT_SIGNATURE: [
    b'\nmV\xb5\xae\x8e\xaf\x17\x16Y\x0e#\xd2lUS\xea\x7fW\xb5',
    b'\x0e\xb8ib\xf8\x9f\xaa\xc4h\xcbuP{Ih\x95B\x84\xbd]'
  ]
}
AUTHORITY_KEY_IDENTIFIERS = [ b'\xca=\xd8\x8e\x0ftW\x7f\xd0\x9a\xd9\xe1!\xbfB\xfb#U)\x86' ]

################################################################################

FLAGS_INTERMEDIATE = 0
FLAGS_FINAL = 1

################################################################################

class Signature:
  def __init__(self, fh):
    def decode_byte(fh):
      b = fh.read(1)
      return ord(b) if b else None
    def decode_leb128(f):
      shift = 0
      val = 0
      while True:
        b = decode_byte(f)
        if b is None:
          return None
        val |= (b & 0x7f) << shift
        if not (b & 0x80):
          break
        shift += 7
      return val
    def decode_bytes(fh):
      n = decode_leb128(fh)
      b = fh.read(n) if n is not None else None
      return b if b else None
    def decode_entry(f):
      return decode_byte(fh), decode_bytes(fh)
    self.version = decode_byte(fh)
    if self.version != 0:
      raise UnsupportedError('Unsupported signature file version')
    self.cert, self.data, self.flags = None, None, None
    while True:
      tag, entry = decode_entry(fh)
      if tag is None:
        break
      if 1 == tag:
        self.cert = entry
      elif 2 == tag:
        self.data = entry
      elif 3 == tag:
        self.flags = entry
    if None in (self.cert, self.data, self.flags):
      raise InvalidDataError('Invalid signature file')
    if ord(self.flags) not in (0, 1):
      raise InvalidDataError('Invalid signature file flags')

################################################################################

def hash_bin(bin):
  with open(bin, 'rb') as fh:
    fmt, p, a = detect_bin(fh)
    if fmt == 'macho':
      return hash_macho_wv0(fh), p, a
    elif fmt == 'pe':
      return hash_pe_wv0(fh), p, a
    raise UnsupportedError('Binary type or architecture not supported')

def load_sig(bin, sig, p):
  if 'win32' == p:
    dir, name = path.split(bin)
    if name != 'electron.exe':
      o = path.join(dir, 'electron.exe.sig')
      if path.isfile(o):
        if not path.isfile(sig):
          rename(o, sig)
        elif not path.samefile(o, sig):
          remove(o)

  try:
    with open(sig, 'rb') as fh:
      return fh.read()
  except Exception as e:
    raise InvalidDataError('Failed to read signature') from e

def verify_dgt(dgt, sdata, kind, intermediate, min_days, any_ski=False, block_size=DEFAULT_HASH_BLOCK):
  try:
    s = Signature(BytesIO(sdata))
  except Exception as e:
    raise InvalidDataError('Failed to decode signature') from e

  expected_flags = FLAGS_INTERMEDIATE if intermediate else FLAGS_FINAL
  if ord(s.flags) != expected_flags:
    raise ValidityError('Signature flags do not match expected value')

  try:
    cert = load_der_x509_certificate(s.cert, backend=default_backend())
  except Exception as e:
    raise InvalidDataError('Failed to load certificate') from e
  try:
    cert.public_key().verify(s.data, dgt + s.flags, PSS(mgf=MGF1(SHA1()), salt_length=20), SHA1())
  except Exception as e:
    raise InvalidDataError('Failed to verify signature') from e

  # Prefer the non-naive UTC attributes if they are available (in cryptography >= 42.0.0)
  if hasattr(cert, 'not_valid_before_utc') and hasattr(cert, 'not_valid_after_utc'):
    now = datetime.now(timezone.utc)
    not_valid_before = cert.not_valid_before_utc
    not_valid_after = cert.not_valid_after_utc
  else:
    now = datetime.now()
    not_valid_before = cert.not_valid_before
    not_valid_after = cert.not_valid_after

  if now < not_valid_before:
    raise ValidityError('Certificate not yet valid')
  if min_days is not None and now > (not_valid_after - timedelta(days=min_days)):
    raise ValidityError(f'Certificate expires in less than {min_days} days')
  days_left = (not_valid_after - now).days

  ski = aki = ku = eku = vp = vd = None
  for e in cert.extensions:
    if ExtensionOID.SUBJECT_KEY_IDENTIFIER == e.oid:
      ski = e.value
    elif ExtensionOID.AUTHORITY_KEY_IDENTIFIER == e.oid:
      aki = e.value
    elif ExtensionOID.KEY_USAGE == e.oid:
      ku = e.value
    elif ExtensionOID.EXTENDED_KEY_USAGE == e.oid:
      eku = e.value
    elif ExtensionOID.VMP_PERSISTENT == e.oid:
      vp = e.value
    elif ExtensionOID.VMP_DEVELOPMENT == e.oid:
      vd = e.value
  if vd:
    raise ValidityError('Certificate is valid for development only')
  if (vp and STREAMING_SIGNATURE == kind) or (not vp and PERSISTENT_SIGNATURE == kind):
    raise ValidityError(f'Certificate doesn\'t match \'{kind}\' signature kind')
  if not eku or ExtendedKeyUsageOID.CODE_SIGNING not in eku:
    raise ValidityError('Certificate doesn\'t allow code signing')
  if not ku or not ku.digital_signature:
    raise ValidityError('Certificate doesn\'t allow digital signatures')
  if not aki or aki.key_identifier not in AUTHORITY_KEY_IDENTIFIERS:
    raise ValidityError('Certificate has invalid AuthorityKeyIdentifier')
  if not any_ski:
    if not ski or ski.digest not in SUBJECT_KEY_IDENTIFIERS.get(kind, []):
      raise ValidityError('Certificate has invalid SubjectKeyIdentifier')
  return (kind, intermediate, days_left)

################################################################################

DEFAULT_CACHE_FILE = '~/.config/evs/cache.json'
DEFAULT_CACHE_MAX = 16

def _cache_key(dgt, p, a, kind, intermediate):
  return f'{p}-{a}:{kind}{"-intermediate" if intermediate else ""}:{b2a_base64(dgt, newline=False).decode("utf-8")}'

################################################################################

MB = 1024 * 1024

################################################################################

class TransferTracker:
  """Tracks bytes processed, elapsed time, retry count, speed, percentage, and outcome.

  success is None while in progress, True on completion, False on failure.
  on_update(tracker) is called after each update().
  on_start(tracker) is called by start() (defaults to on_update).
  on_done(tracker) is called by done() (defaults to on_update).

  input_size counts all bytes sent over the wire (monotonically increasing).
  output_size counts resulting bytes, this can decrease e.g. in case of retries.
  input_pct reflects the input position (capped at 100%).
  output_pct reflects the output position (capped at 100%).
  speed reflects total wire throughput.
  retries reflects the number of times a negative progress update was detected.
  """
  def __init__(self, size=0, on_update=None, on_start=None, on_done=None):
    self._size = size
    self._start = None
    self._input_size = 0
    self._retries = 0
    self._output_size = 0
    self.success = None
    self._on_update = on_update
    self._on_start = on_start if on_start is not None else on_update
    self._on_done = on_done if on_done is not None else on_update

  def __enter__(self):
    self.start()
    return self

  def __exit__(self, exc_type, exc_value, tb):
    self.done(exc_type is None)

  def start(self):
    assert self._start is None
    assert self.success is None
    self._start = monotonic()
    if self._on_start:
      self._on_start(self)

  def update(self, input_size, output_size):
    assert self.start is not None
    assert self.success is None
    if input_size is not None:
      self._input_size = input_size
    if output_size is not None:
      if output_size < self._output_size:
        self._retries += 1
      self._output_size = output_size
    if self._on_update:
      self._on_update(self)

  def update_input_n(self, n):
    assert n >= 0
    self.update(self._input_size + n, None)
    return n

  def update_output_n(self, n):
    self.update(None, self._output_size + n)
    return n

  def update_n(self, n):
    if n < 0:
      return self.update_output_n(n)
    self.update(self._input_size + n, self._output_size + n)
    return n

  def done(self, success=True):
    assert self.success is None
    self.success = success
    if self._on_done:
      self._on_done(self)

  @property
  def size(self):
    return self._size

  @property
  def elapsed(self):
    return (monotonic() - self._start) if self._start else 0

  @property
  def input_size(self):
    return self._input_size

  @property
  def output_size(self):
    return self._output_size

  @property
  def input_pct(self):
    return min(100.0, self.input_size / self._size * 100) if self._size else 0.0

  @property
  def output_pct(self):
    return min(100.0, self.output_size / self._size * 100) if self._size else 0.0

  @property
  def speed(self):
    e = self.elapsed
    return self._input_size / e / MB if e > 0 else 0.0

  @property
  def retries(self):
    return self._retries

################################################################################

class DataProgressTracker(TransferTracker):
  """Wraps file-like object for reading, and optionally another for writing, for progress tracking.

  compatible with the with statement for automatic management of start/done states.
  tracks bytes read/written and signals updates via callback.
  finalize() can be called to retreive the optional output buffer and size.
  """
  def __init__(self, in_size, in_data, out_data=None, on_update=None, on_start=None, on_done=None):
    super().__init__(in_size, on_update=on_update, on_start=on_start, on_done=on_done)
    self._in_data = in_data
    self._out_data = out_data

  def __len__(self):
    return self.size

  def read(self, size=-1):
    chunk = self._in_data.read(size)
    if chunk:
      n = len(chunk)
      self.update_n(n) if self._out_data is None else self.update_input_n(n)
    return chunk

  def write(self, data):
    return self.update_output_n(self._out_data.write(data))

  def finalize(self):
    out_size = self.output_size
    if self._out_data is None:
      return (None, out_size)
    out_data = self._out_data
    assert out_size == out_data.tell()
    self._out_data = None
    out_data.seek(0)
    return (out_data, out_size)

################################################################################

class UploadSubscriberTracker(TransferTracker):
  """s3transfer subscriber for accurate per-byte progress and retry tracking.

  on_progress() receives negative bytes_transferred when s3transfer seeks back to retry a part.
  update_n() accumulates those as output_size, dropping pct back to the confirmed position
  while input_size keeps growing for speed.
  """
  def __init__(self, size, on_update=None, on_start=None, on_done=None):
    super().__init__(size, on_update=on_update, on_start=on_start, on_done=on_done)

  def on_queued(self, future, **kwargs):
    pass

  def on_progress(self, future, bytes_transferred, **kwargs):
    self.update_n(bytes_transferred)

  def on_done(self, future, **kwargs):
    pass

################################################################################

class LogFormatter:
  """TTY-aware string formatter for log output."""

  def __init__(self, log_fn, tty=None):
    self._tty = tty if tty is not None else stdout.isatty()
    self._log_fn = log_fn
    if self._tty:
      self._br_open  = '\033[36m[ \033[0m'
      self._br_close = '\033[36m ]\033[0m'
      self._sep      = '\033[36m | \033[0m'
    else:
      self._br_open  = '[ '
      self._br_close = ' ]'
      self._sep      = ' | '

  def _bold(self, s):
    return f'\033[1m{s}\033[0m' if self._tty else s

  def log_bold(self, prefix, s):
    self._log_fn(f'{prefix} {self._bold(s)}')

  def _log_progress(self, t, prefix, output_pct=True):
    msg = f'\r{prefix} {self._br_open}{t.input_pct if not output_pct else t.output_pct:.1f}%'
    if t.retries:
      msg += f'{self._sep}{t.retries} retr{"y" if t.retries == 1 else "ies"}'
    msg += f'{self._sep}{t.input_size / MB:.1f} of {t.size / MB:.1f} MB'
    if t.output_size != t.input_size:
      msg += f' \u2192 {t.output_size / MB:.1f} MB'
    msg += f'{self._sep}{t.speed:.1f} MB/s{self._br_close}\033[K'
    self._log_fn(msg, end='')

  def log_transfer_start(self, t, prefix, output_pct=True):
    if self._tty:
      self._log_progress(t, prefix, output_pct)
    else:
      self._log_fn(f'{prefix} {t.size / MB:.1f} MB')

  def log_transfer_update(self, t, prefix, output_pct=True):
    if self._tty:
      self._log_progress(t, prefix, output_pct)

  def log_transfer_done(self, t, prefix_fmt, parts=None):
    if self._tty:
      self._log_fn('')
    msg = prefix_fmt.format('finished' if t.success else 'failed')
    if parts is not None:
      msg += f' {self._bold(f"{parts}")}'
      msg += f' part{"s" if parts != 1 else ""},'
    if t.retries:
      msg += f' {self._bold(f"{t.retries}")}'
      msg += f' retr{"y" if t.retries == 1 else "ies"},'
    if t.input_size != t.size:
      msg += f' {self._bold(f"{t.input_size / MB:.1f}")} of'
    msg += f' {self._bold(f"{t.size / MB:.1f}")} MB'
    if t.output_size not in (t.input_size, t.size):
      msg += f' \u2192 {self._bold(f"{t.output_size / MB:.1f}")} MB'
    msg += f', {self._bold(f"{t.elapsed:.1f}s")}, avg {self._bold(f"{t.speed:.1f}")} MB/s'
    self._log_fn(msg)

  def log_signature(self, prefix, kind, intermediate, days_left):
    msg = f'{prefix} {self._bold(kind)}'
    if intermediate:
      msg += f', {self._bold("intermediate")}'
    if days_left is not None:
      msg += f', {self._bold(f"{days_left}")} days left'
    self._log_fn(msg)

################################################################################

class VMP(Auth):
  DEFAULT_UPLOAD_URL_TIMEOUT = 60
  DEFAULT_UPLOAD_TIMEOUT = 300
  DEFAULT_SIGN_TIMEOUT = 120

  MULTIPART_MIN_SIZE = 20            # MB — minimum for threshold and part size
  DEFAULT_MULTIPART_THRESHOLD = 100  # MB
  DEFAULT_MULTIPART_PART_SIZE = 50   # MB
  DEFAULT_MULTIPART_RETRIES = 3
  DEFAULT_MULTIPART_MAX_CONCURRENCY = 2

  def __init__(self, no_ask=True, any_ski=False, api_url=None, cache_file=None, cache_max=None,
               connect_timeout=None, auth_timeout=None, upload_url_timeout=None, upload_timeout=None,
               sign_timeout=None, multipart_threshold=None, multipart_part_size=None,
               multipart_retries=None, multipart_max_concurrency=None, **_):
    self._any_ski = any_ski or 'EVS_ANY_SKI' in environ
    self._cache_file = path.expanduser(cache_file or environ.get('EVS_CACHE_FILE', DEFAULT_CACHE_FILE))
    self._cache_dir = path.dirname(self._cache_file)
    self._cache_max = int(cache_max or environ.get('EVS_CACHE_MAX', DEFAULT_CACHE_MAX))
    self._upload_url_timeout = float(upload_url_timeout or environ.get('EVS_UPLOAD_URL_TIMEOUT', self.DEFAULT_UPLOAD_URL_TIMEOUT))
    self._upload_timeout = float(upload_timeout or environ.get('EVS_UPLOAD_TIMEOUT', self.DEFAULT_UPLOAD_TIMEOUT))
    self._sign_timeout = float(sign_timeout or environ.get('EVS_SIGN_TIMEOUT', self.DEFAULT_SIGN_TIMEOUT))
    multipart_threshold = int(multipart_threshold or environ.get('EVS_MULTIPART_THRESHOLD', self.DEFAULT_MULTIPART_THRESHOLD))
    multipart_part_size = int(multipart_part_size or environ.get('EVS_MULTIPART_PART_SIZE', self.DEFAULT_MULTIPART_PART_SIZE))
    self._multipart_threshold = max(self.MULTIPART_MIN_SIZE, multipart_threshold) * MB
    self._multipart_part_size = max(self.MULTIPART_MIN_SIZE, multipart_part_size) * MB
    self._multipart_retries = int(multipart_retries or environ.get('EVS_MULTIPART_RETRIES', self.DEFAULT_MULTIPART_RETRIES))
    self._multipart_max_concurrency = int(multipart_max_concurrency or environ.get('EVS_MULTIPART_MAX_CONCURRENCY', self.DEFAULT_MULTIPART_MAX_CONCURRENCY))
    api_url = api_url or environ.get('EVS_API_URL', DEFAULT_API_URL)
    self.broker_url = f'{api_url}/broker'
    self.signer_url = f'{api_url}/signer'
    super().__init__(no_ask=no_ask, connect_timeout=connect_timeout, auth_timeout=auth_timeout)
    self._fmt = LogFormatter(self._log)
    self._agent = f'{__name__}-{self._version}'
    self._load_cache()

  def _load_cache(self):
    try:
      with open(self._cache_file) as fh:
        self.__cache = json.load(fh)
    except Exception:
      self.__cache = {}

  def _compress(self, data, size, chunk_size=1 << 20):
    p0, p1 = ' - Compressing binary:', ' - Compression {}:'
    with DataProgressTracker(size, data, BytesIO(),
        on_update=lambda t: self._fmt.log_transfer_update(t, p0, False),
        on_start=lambda t: self._fmt.log_transfer_start(t, p0, False),
        on_done=lambda t: self._fmt.log_transfer_done(t, p1)) as tracked_io_data:
      cobj = compressobj(wbits=MAX_WBITS | 16)
      while True:
        chunk = tracked_io_data.read(chunk_size)
        if not chunk:
          break
        tracked_io_data.write(cobj.compress(chunk))
      tracked_io_data.write(cobj.flush())
      return tracked_io_data.finalize()

  def _store_cache(self):
    if not self._cache_file:
      return False
    try:
      try:
        makedirs(self._cache_dir)
      except OSError as e:
        if errno.EEXIST != e.errno or not path.isdir(self._cache_dir):
          return False
      with open(self._cache_file, 'w') as fh:
        json.dump(self.__cache, fh, indent=2)
      return True
    except Exception:
      return False

  def _pop_cache_item(self, key):
    sdata = self.__cache.pop(key, None)
    if not sdata:
      return None
    return a2b_base64(sdata)

  def _add_cache_item(self, key, sdata):
    # Delete before re-inserting to refresh insertion order for LRU eviction
    if key in self.__cache:
      del self.__cache[key]
    self.__cache[key] = b2a_base64(sdata, newline=False).decode('utf-8')
    while len(self.__cache) > self._cache_max:
      del self.__cache[next(iter(self.__cache))]

  def _request_json(self, url, params, timeout, desc, account_name=None, passwd=None):
    try:
      while True:
        r = requests.get(url,
          params=params,
          headers={
            'Authorization': self._auth_id_token,
            'Accept': 'application/json',
          },
          timeout=(self._connect_timeout, timeout),
        )
        if r.status_code not in (401, 403) or not self._refresh_auth(account_name, passwd):
          break
    except Exception as e:
      raise RequestError(f'{desc} failed') from e
    try:
      j = r.json()
    except Exception as e:
      raise InvalidDataError(f'{desc} failed: Could not parse JSON response') from e
    if r.status_code != 200 or 'errorMessage' in j or 'message' in j:
      e = gather(j)
      raise HTTPError(f'{desc} failed: {r.status_code} {r.reason}') from e
    return j

  def _request_upload(self, gz=False, account_name=None, passwd=None):
    ps = {
      'agent': self._agent,
      **({'gz': ''} if gz else {})
    }
    j = self._request_json(self.broker_url, ps, self._upload_url_timeout,
      'Request for upload URL', account_name, passwd)
    if 'key' not in j:
      raise InvalidDataError('No \'key\' returned in upload URL response')
    if 'url' not in j:
      raise InvalidDataError('No \'url\' returned in upload URL response')
    return (j['key'], j['url'])

  def _upload_single(self, data, size, gz=False, account_name=None, passwd=None):
    self._fmt.log_bold(' - Acquiring upload path for mode:', 'single')
    key, url = self._request_upload(gz, account_name, passwd)
    if key.endswith('.gz') != gz:
      raise InvalidDataError('Unexpected gz status mismatch in upload key')
    self._fmt.log_bold(' - Acquired upload path:', key)
    p0, p1 = ' - Uploading binary:', ' - Upload {}:'
    with DataProgressTracker(size, data,
        on_update=lambda t: self._fmt.log_transfer_update(t, p0),
        on_start=lambda t: self._fmt.log_transfer_start(t, p0),
        on_done=lambda t: self._fmt.log_transfer_done(t, p1)) as tracked_data:
      try:
        r = requests.put(url,
          data=tracked_data,
          headers={
            'Content-Type': 'application/octet-stream' if not gz else 'application/gzip',
          },
          timeout=(self._connect_timeout, self._upload_timeout),
        )
      except Exception as e:
        raise RequestError('Upload request failed') from e
      if r.status_code != 200:
        raise HTTPError(f'Upload request failed: {r.status_code} {r.reason}')
      return key

  def _request_multipart_upload(self, gz=False, account_name=None, passwd=None):
    ps = {'agent': self._agent, 'multipart': '', **({'gz': ''} if gz else {})}
    j = self._request_json(self.broker_url, ps, self._upload_url_timeout,
      'Request for multipart upload credentials', account_name, passwd)
    for field in ('key', 'bucket', 'region', 'credentials'):
      if field not in j:
        raise InvalidDataError(f'No \'{field}\' returned in multipart upload response')
    return j['key'], j['bucket'], j['region'], j['credentials']

  def _upload_multipart(self, data, size, gz=False, account_name=None, passwd=None):
    self._fmt.log_bold(' - Acquiring upload path for mode:', 'multipart')
    key, bucket, region, creds = self._request_multipart_upload(gz, account_name, passwd)
    if key.endswith('.gz') != gz:
      raise InvalidDataError('Unexpected gz status mismatch in multipart upload key')
    self._fmt.log_bold(' - Acquired upload path:', key)
    parts = ceil(size / self._multipart_part_size)
    p0, p1 = ' - Uploading binary:', ' - Upload {}:'
    with UploadSubscriberTracker(size,
        on_update=lambda t: self._fmt.log_transfer_update(t, p0),
        on_start=lambda t: self._fmt.log_transfer_start(t, p0),
        on_done=lambda t: self._fmt.log_transfer_done(t, p1, parts)) as upload_tracker:
      s3 = boto3.client('s3',
        region_name=region,
        aws_access_key_id=creds['access_key_id'],
        aws_secret_access_key=creds['secret_access_key'],
        aws_session_token=creds['session_token'],
        config=Config(retries={'max_attempts': self._multipart_retries, 'mode': 'adaptive'}),
      )
      tc = TransferConfig(
        multipart_threshold=1,
        multipart_chunksize=self._multipart_part_size,
        max_concurrency=self._multipart_max_concurrency,
        # CRT client provides worse progress and error handling for our case, use classic client
        preferred_transfer_client='classic',
      )
      with create_transfer_manager(s3, tc) as manager:
        manager.upload(data, bucket, key,
          extra_args={'ContentType': 'application/gzip' if gz else 'application/octet-stream'},
          subscribers=[upload_tracker]
        ).result()
      return key

  def _upload_data(self, data, gz=False, account_name=None, passwd=None):
    data.seek(0, 2)
    size = data.tell()
    data.seek(0)
    if gz:
      data, size = self._compress(data, size)
    if size >= self._multipart_threshold:
      return self._upload_multipart(data, size, gz, account_name, passwd)
    return self._upload_single(data, size, gz, account_name, passwd)

  def _upload_file(self, bin, gz=False, account_name=None, passwd=None):
    with open(bin, 'rb') as fh:
      return self._upload_data(fh, gz, account_name, passwd)

  def _sign(self, key, kind, intermediate=False, account_name=None, passwd=None):
    ps = {
      'agent': self._agent,
      'key': key,
      'kind': kind,
      **({'intermediate': ''} if intermediate else {})
    }
    j = self._request_json(self.signer_url, ps, self._sign_timeout,
      'Signing request', account_name, passwd)
    if 'key' not in j:
      raise InvalidDataError('No \'key\' returned in signing response')
    if key != j['key']:
      raise InvalidDataError('Mismatching \'key\' returned in signing response')
    if 'sig' not in j:
      raise InvalidDataError('No \'sig\' returned in signing response')
    return a2b_base64(j['sig'])

  def _store(self, sig, sdata):
    with open(sig, 'wb') as fh:
      fh.write(sdata)

  def verify(self, bin, sig=None, kind=None, intermediate=False, min_days=None, app=None, **_):
    sig = sig or (bin + '.sig')
    kind = kind or environ.get('EVS_KIND', STREAMING_SIGNATURE)
    intermediate = intermediate or 'EVS_INTERMEDIATE' in environ
    self._fmt.log_bold('Verifying:', app or bin)
    dgt, p, _ = hash_bin(bin)
    msg = verify_dgt(dgt, load_sig(bin, sig, p), kind, intermediate, min_days, self._any_ski)
    self._fmt.log_signature(' - Signature is valid:', *msg)

  def sign(self, bin, sig=None, kind=None, intermediate=False, min_days=None, force=False, gz=False, account_name=None, passwd=None, app=None, **_):
    sig = sig or (bin + '.sig')
    key = environ.get('EVS_KEY', None)
    kind = kind or environ.get('EVS_KIND', STREAMING_SIGNATURE)
    intermediate = intermediate or 'EVS_INTERMEDIATE' in environ
    force = force or 'EVS_FORCE' in environ
    self._fmt.log_bold('Signing:', app or bin)
    dgt, p, a = hash_bin(bin)
    ckey = _cache_key(dgt, p, a, kind, intermediate)
    sdata = None
    if not force:
      self._log(' - Checking existing signature')
      try:
        sdata = load_sig(bin, sig, p)
        msg = verify_dgt(dgt, sdata, kind, intermediate, min_days, self._any_ski)
        self._fmt.log_signature(' - Retain existing signature:', *msg)
        return
      except Exception as e:
        self._log(f' - Existing signature invalid: {e}')
      sdata = self._pop_cache_item(ckey)
      if sdata:
        self._log(' - Checking cached signature')
        try:
          msg = verify_dgt(dgt, sdata, kind, intermediate, min_days, self._any_ski)
          self._fmt.log_signature(' - Using cached signature:', *msg)
        except Exception as e:
          self._log(f' - Cached signature invalid: {e}')
          sdata = None
      else:
        self._log(' - No cached signature found')
    if not sdata:
      self._check_auth(account_name, passwd)
      key = key or self._upload_file(bin, gz, account_name, passwd)
      self._fmt.log_signature(' - Requesting signature:', kind, intermediate, None)
      sdata = self._sign(key, kind, intermediate, account_name, passwd)
      try:
        msg = verify_dgt(dgt, sdata, kind, intermediate, min_days, self._any_ski)
        self._fmt.log_signature(' - Signature request successful:', *msg)
      except Exception as e:
        raise Exception(f'Signature verification failed after signing: {e}') from e
    self._add_cache_item(ckey, sdata)
    self._store_cache()
    self._store(sig, sdata)

  def _detect_pkgs(self, dirs, name_hint=None):
    def _to_pattern(n):
      return ''.join((f'[{c.upper()}{c.lower()}]' if c.isalpha() else c) for c in n)
    def _try_app(app):
      if path.isdir(app):
        fwdir = 'Contents/Frameworks/Electron Framework.framework/Versions/A'
        fwbin = path.join(app, fwdir, 'Electron Framework')
        if path.isfile(fwbin):
          fwsig = path.join(app, fwdir, 'Resources/Electron Framework.sig')
          return (app, fwbin, fwsig)
      return None
    def _try_exe(exe):
      if path.isfile(exe):
        sig = f'{exe}.sig'
        return (exe, exe, sig)
      return None
    def _try_dir(dir, patterns):
      for p in patterns:
        p = path.join(dir, p)
        for m in glob(f'{p}.app'):
          item = _try_app(m)
          if item: return item
        for m in glob(f'{p}.exe'):
          item = _try_exe(m)
          if item: return item
      raise FileNotFoundError(f'No matching executable found in: {dir}')

    patterns = []
    if not name_hint:
      patterns.append(_to_pattern('electron'))
      patterns.append('*')
    else:
      patterns.append(_to_pattern(name_hint))
    return [ _try_dir(dir, patterns) for dir in dirs ]

  def verify_pkg(self, pkg, name_hint=None, kind=None, min_days=None, **_):
    items = self._detect_pkgs(pkg if isinstance(pkg, list) else [ pkg ], name_hint)
    for item in items:
      self.verify(item[1], item[2], kind, False, min_days, app=item[0])

  def sign_pkg(self, pkg, name_hint=None, kind=None, min_days=None, force=False, gz=False, account_name=None, passwd=None, **_):
    items = self._detect_pkgs(pkg if isinstance(pkg, list) else [ pkg ], name_hint)
    for item in items:
      self.sign(item[1], item[2], kind, False, min_days, force, gz, account_name, passwd, app=item[0])

################################################################################

if __name__ == '__main__':
  from .cli.vmp import main
  main()

################################################################################



# Copyright 2020, castLabs GmbH

from .error import UnsupportedError, InvalidDataError

from sys import maxsize

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.hashes import Hash, SHA512, SHA1

from macholib import MachO, mach_o

################################################################################

def detect_bin(fh):
  h = fh.peek(512)
  if h[:8] == b'\xcf\xfa\xed\xfe\x07\x00\x00\x01':
    return ('macho', 'darwin', 'x64')
  elif h[:8] == b'\xcf\xfa\xed\xfe\x0C\x00\x00\x01':
    return ('macho', 'darwin', 'arm64')
  elif h[:4] == b'\xca\xfe\xba\xbe':
    return ('macho', 'darwin', 'fat')
  elif h[:2] == b'MZ':
    o = int.from_bytes(h[60:64], byteorder='little')
    if o < len(h):
      s = h[o:o + 6]
      if s == b'\x50\x45\x00\x00\x64\x86':
        return ('pe', 'win32', 'x64')
      elif s == b'\x50\x45\x00\x00\x4c\x01':
        return ('pe', 'win32', 'ia32')
      elif s == b'\x50\x45\x00\x00\x64\xaa':
        return ('pe', 'win32', 'arm64')
  raise UnsupportedError('Binary type or architecture not supported')

################################################################################

DEFAULT_HASH_BLOCK = 2 * 1024 * 1024

################################################################################

def hash_macho_wv0(fh, hash_block=DEFAULT_HASH_BLOCK):
  headers = MachO.MachO(fh.name).headers
  hasher = Hash(SHA512(), default_backend())
  for header in headers:
    fh.seek(header.offset, 0)
    start, end = maxsize, 0
    for (lc, segment, sections) in header.commands:
      # The minimum section offset of all load commands is the start of VMP signing part
      if (lc.cmd in (mach_o.LC_SEGMENT_64, mach_o.LC_SEGMENT) and
        segment.segname.startswith(mach_o.SEG_TEXT.encode('utf-8'))):
        for section in sections:
          start = min(start, section.offset)
      # Expect the String Table is at the end of unsigned binary followed by the code
      # signature, so the end of String Table is the end of VMP signing part
      if (mach_o.LC_SYMTAB == lc.cmd):
        end = segment.stroff + segment.strsize
    if (start >= end):
      raise InvalidDataError('Failed to assemble VMP/Mach-O signing body: %d-%d' % (start, end))
    fh.seek(start, 1)
    while start < end:
      data = fh.read(min(end - start, hash_block))
      start += len(data)
      hasher.update(data)
  fh.seek(0, 0)
  return hasher.finalize()

def hash_macho_fat_wv0(fh, hash_block=DEFAULT_HASH_BLOCK):
  headers = MachO.MachO(fh.name).headers
  hashes = []
  for header in headers:
    fh.seek(header.offset, 0)
    start, end = maxsize, 0
    for (lc, segment, sections) in header.commands:
      # The minimum section offset of all load commands is the start of VMP signing part
      if (lc.cmd in (mach_o.LC_SEGMENT_64, mach_o.LC_SEGMENT) and
        segment.segname.startswith(mach_o.SEG_TEXT.encode('utf-8'))):
        for section in sections:
          start = min(start, section.offset)
      # Expect the String Table is at the end of unsigned binary followed by the code
      # signature, so the end of String Table is the end of VMP signing part
      if (mach_o.LC_SYMTAB == lc.cmd):
        end = segment.stroff + segment.strsize
    if (start >= end):
      raise InvalidDataError('Failed to assemble VMP/Mach-O signing body: %d-%d' % (start, end))
    fh.seek(start, 1)
    hasher = Hash(SHA512(), default_backend())
    while start < end:
      data = fh.read(min(end - start, hash_block))
      start += len(data)
      hasher.update(data)
    hashes.append(hasher.finalize())
  fh.seek(0, 0)
  return hashes

def hash_pe_wv0(fh, hash_block=DEFAULT_HASH_BLOCK):
  hasher = Hash(SHA512(), default_backend())
  while True:
    data = fh.read(hash_block)
    if not data: break
    hasher.update(data)
  fh.seek(0, 0)
  return hasher.finalize()

################################################################################



# Copyright 2020, castLabs GmbH

from .. import __version__
from ..core.failure import die
from ..vmp import VMP
from argparse import ArgumentParser
from sys import argv
from os import environ, path

################################################################################

def main():
  try:
    bp = ArgumentParser(path.basename(argv[0]))
    bp.add_argument('-v', '--version', action='version', version=f'%(prog)s {__version__}')
    bp.add_argument('-n', '--no-ask', help='enable non-interactive mode',
      action='store_true', default='EVS_NO_ASK' in environ
    )
    bp.add_argument('-a', '--any-ski', help='verify any subject key identifier',
      action='store_true', default='EVS_ANY_SKI' in environ
    )

    _kind = ArgumentParser(add_help=False)
    _kind.set_defaults(kind=environ.get('EVS_KIND', 'streaming'))
    _kind.add_argument('-s', '--streaming', help='streaming only signature (default)',
      dest='kind', action='store_const', const='streaming'
    )
    _kind.add_argument('-p', '--persistent', help='streaming and persistent download signature',
      dest='kind', action='store_const', const='persistent'
    )
    _flags = ArgumentParser(add_help=False)
    _flags.add_argument('-i', '--intermediate', help='generate intermediate signature',
      dest='intermediate', action='store_true', default='EVS_INTERMEDIATE' in environ
    )
    _min_days = ArgumentParser(add_help=False)
    _min_days.add_argument('-M', '--min-days', help='minimum number of remaining valid days',
      type=int, default=environ.get('EVS_MIN_DAYS', 90)
    )
    _force = ArgumentParser(add_help=False)
    _force.add_argument('-f', '--force', help='force new signature',
      action='store_true', default='EVS_FORCE' in environ
    )
    _gz = ArgumentParser(add_help=False)
    _gz.add_argument('-z', '--gz', help='request gzip upload (slower for fast connections)',
      action='store_true', default='EVS_GZ' in environ
    )
    _cred = ArgumentParser(add_help=False)
    _cred.add_argument('-A', '--account-name', help='provide account name [CACHED]',
      default=environ.get('EVS_ACCOUNT_NAME')
    )
    _cred.add_argument('-P', '--passwd', help='provide account password',
      default=environ.get('EVS_PASSWD')
    )
    _transfer = ArgumentParser(add_help=False)
    _transfer_grp = _transfer.add_argument_group('transfer options')
    _transfer_grp.add_argument('--connect-timeout',
      help=f'connection timeout in seconds (default: {VMP.DEFAULT_CONNECT_TIMEOUT})',
      type=float, default=environ.get('EVS_CONNECT_TIMEOUT', VMP.DEFAULT_CONNECT_TIMEOUT)
    )
    _transfer_grp.add_argument('--auth-timeout',
      help=f'auth request timeout in seconds (default: {VMP.DEFAULT_AUTH_TIMEOUT})',
      type=float, default=environ.get('EVS_AUTH_TIMEOUT', VMP.DEFAULT_AUTH_TIMEOUT)
    )
    _transfer_grp.add_argument('--upload-url-timeout',
      help=f'upload URL request timeout in seconds (default: {VMP.DEFAULT_UPLOAD_URL_TIMEOUT})',
      type=float, default=environ.get('EVS_UPLOAD_URL_TIMEOUT', VMP.DEFAULT_UPLOAD_URL_TIMEOUT)
    )
    _transfer_grp.add_argument('--upload-timeout',
      help=f'upload request timeout in seconds (default: {VMP.DEFAULT_UPLOAD_TIMEOUT})',
      type=float, default=environ.get('EVS_UPLOAD_TIMEOUT', VMP.DEFAULT_UPLOAD_TIMEOUT)
    )
    _transfer_grp.add_argument('--sign-timeout',
      help=f'sign request timeout in seconds (default: {VMP.DEFAULT_SIGN_TIMEOUT})',
      type=float, default=environ.get('EVS_SIGN_TIMEOUT', VMP.DEFAULT_SIGN_TIMEOUT)
    )
    _transfer_grp.add_argument('--multipart-threshold',
      help=f'multipart upload threshold in MB (default: {VMP.DEFAULT_MULTIPART_THRESHOLD}, min: {VMP.MULTIPART_MIN_SIZE})',
      type=int, default=environ.get('EVS_MULTIPART_THRESHOLD', VMP.DEFAULT_MULTIPART_THRESHOLD)
    )
    _transfer_grp.add_argument('--multipart-part-size',
      help=f'multipart upload part size in MB (default: {VMP.DEFAULT_MULTIPART_PART_SIZE}, min: {VMP.MULTIPART_MIN_SIZE})',
      type=int, default=environ.get('EVS_MULTIPART_PART_SIZE', VMP.DEFAULT_MULTIPART_PART_SIZE)
    )
    _transfer_grp.add_argument('--multipart-retries',
      help=f'multipart upload max retries (default: {VMP.DEFAULT_MULTIPART_RETRIES})',
      type=int, default=environ.get('EVS_MULTIPART_RETRIES', VMP.DEFAULT_MULTIPART_RETRIES)
    )
    _transfer_grp.add_argument('--multipart-max-concurrency',
      help=f'multipart upload max concurrent connections (default: {VMP.DEFAULT_MULTIPART_MAX_CONCURRENCY})',
      type=int, default=environ.get('EVS_MULTIPART_MAX_CONCURRENCY', VMP.DEFAULT_MULTIPART_MAX_CONCURRENCY)
    )

    _name_hint = ArgumentParser(add_help=False)
    _name_hint.add_argument('-H', '--name-hint', help='provide name hint',
      default=environ.get('EVS_NAME_HINT')
    )

    _assets = ArgumentParser(add_help=False)
    _assets.add_argument('bin', help='path to binary')
    _assets.add_argument('sig', help='path to signature', nargs='?')
    _pkgs = ArgumentParser(add_help=False)
    _pkgs.add_argument('pkg', help='path to electron package directory', nargs='+')

    sp = bp.add_subparsers(title='commands', dest='command', required=True)
    cp = sp.add_parser('verify', help='verify signature',
      aliases=['v'],
      parents=[_kind, _flags, _min_days, _assets]
    )
    cp.set_defaults(command=VMP.verify)
    cp = sp.add_parser('sign', help='refresh signature',
      aliases=['s'],
      parents=[_kind, _flags, _min_days, _force, _gz, _cred, _transfer, _assets]
    )
    cp.set_defaults(command=VMP.sign)
    cp = sp.add_parser('verify-pkg', help='verify electron package signature',
      aliases=['vp'],
      parents=[_kind, _name_hint, _min_days, _pkgs]
    )
    cp.set_defaults(command=VMP.verify_pkg)
    cp = sp.add_parser('sign-pkg', help='refresh electron package signature',
      aliases=['sp'],
      parents=[_kind, _name_hint, _min_days, _force, _gz, _cred, _transfer, _pkgs]
    )
    cp.set_defaults(command=VMP.sign_pkg)

    args = vars(bp.parse_args())
    vmp = VMP(no_ask=args.pop('no_ask'), any_ski=args.pop('any_ski'),
              connect_timeout=args.pop('connect_timeout', None),
              auth_timeout=args.pop('auth_timeout', None),
              upload_url_timeout=args.pop('upload_url_timeout', None),
              upload_timeout=args.pop('upload_timeout', None),
              sign_timeout=args.pop('sign_timeout', None),
              multipart_threshold=args.pop('multipart_threshold', None),
              multipart_part_size=args.pop('multipart_part_size', None),
              multipart_retries=args.pop('multipart_retries', None),
              multipart_max_concurrency=args.pop('multipart_max_concurrency', None))
    run = args.pop('command')
    run(vmp, **args)
  except (Exception, KeyboardInterrupt) as e:
    die(e)

################################################################################
